@@ -333,7 +333,7 @@ static TrackerExtractData extract_data[] = {
 	{ NULL, NULL }
 };
 
-static void
+static gchar *
 improve_handwritten_genre (gchar *genre)
 {
 	/* This function tries to make each first letter of each word
@@ -341,25 +341,30 @@ improve_handwritten_genre (gchar *genre)
 	 * example, if it is "Fusion jazz", we want "Fussion Jazz" to
 	 * make things more consistent.
 	 */
-        gchar *p;
-	gunichar c;
+
+	gunichar *str;
+	gchar    *ret;
+
+        gunichar *p;
 	gboolean set_next;
 
 	if (!genre) {
-		return;
+		return NULL;
 	}
 
-	c = g_utf8_get_char (genre);
-	*genre = g_unichar_toupper (c);
+	str = g_utf8_to_ucs4 (genre, -1, NULL, NULL, NULL);
 
-        for (p = genre, set_next = FALSE; *p; p = g_utf8_next_char (p)) {
+	if (!str) {
+		return NULL;
+	}
+
+        for (p = str, set_next = TRUE; *p; p++) {
 		GUnicodeBreakType t;
 
-                c = g_utf8_get_char (p);
-		t = g_unichar_break_type (c);
+		t = g_unichar_break_type (*p);
 
 		if (set_next) {
-			*p = g_unichar_toupper (c);
+			*p = g_unichar_toupper (*p);
 			set_next = FALSE;
 		}
 
@@ -405,6 +410,13 @@ improve_handwritten_genre (gchar *genre)
 			break;
 		}
         }
+
+	ret = g_ucs4_to_utf8 (str, -1, NULL, NULL, NULL);
+
+	g_free (str);
+
+	return ret;
+
 }
 
 static char *
@@ -1056,14 +1068,14 @@ get_id3v24_tags (const gchar *data,
 							  NULL, NULL, NULL);
 					break;
 				case 0x01 :
-					word = t_convert (&data[pos + 11],
+					word = g_convert (&data[pos + 11],
 							  csize - 1,
 							  "UTF-8",
 							  info->encoding ? info->encoding : "UTF-16",
 							  NULL, NULL, NULL);
 					break;
 				case 0x02 :
-					word = t_convert (&data[pos + 11],
+					word = g_convert (&data[pos + 11],
 							  csize - 1,
 							  "UTF-8",
 							  info->encoding ? info->encoding : "UTF-16BE",
@@ -1100,6 +1112,7 @@ get_id3v24_tags (const gchar *data,
 						g_strfreev (parts);
 					} else if (strcmp (tmap[i].text, "TCON") == 0) {
 						gint genre;
+						gchar *improved_genre;
 
 						if (get_genre_number (word, &genre)) {
 							g_free (word);
@@ -1108,9 +1121,12 @@ get_id3v24_tags (const gchar *data,
 							if (g_ascii_strcasecmp (word, "unknown") == 0) {
 								g_free (word);
 								break;
-							} 
-
-							improve_handwritten_genre (word);
+							} else {
+								if ((improved_genre = improve_handwritten_genre (word)) != NULL) {
+									g_free (word);
+									word = improved_genre;
+								}
+							}
 						}
 					} else if (strcmp (tmap[i].text, "TLEN") == 0) {
 						guint32 duration;
@@ -1168,14 +1184,14 @@ get_id3v24_tags (const gchar *data,
 						  NULL, NULL, NULL);
 				break;
 			case 0x01:
-				word = t_convert (text,
+				word = g_convert (text,
 						  csize - offset,
 						  "UTF-8",
 						  info->encoding ? info->encoding : "UTF-16",
 						  NULL, NULL, NULL);
 				break;
 			case 0x02:
-				word = t_convert (text,
+				word = g_convert (text,
 						  csize-offset,
 						  "UTF-8",
 						  info->encoding ? info->encoding : "UTF-16BE",
@@ -1362,6 +1378,7 @@ get_id3v23_tags (const gchar *data,
 						g_strfreev (parts);
 					} else if (strcmp (tmap[i].text, "TCON") == 0) {
 						gint genre;
+						gchar *improved_genre;
 
 						if (get_genre_number (word, &genre)) {
 							g_free (word);
@@ -1370,9 +1387,12 @@ get_id3v23_tags (const gchar *data,
 							if (g_ascii_strcasecmp (word, "unknown") == 0) {
 								g_free (word);
 								break;
+							} else {
+								if ((improved_genre = improve_handwritten_genre (word)) != NULL) {
+									g_free (word);
+									word = improved_genre;
+								}
 							} 
-
-							improve_handwritten_genre (word);
 						}
 					} else if (strcmp (tmap[i].text, "TLEN") == 0) {
 						guint32 duration;
@@ -1528,6 +1548,7 @@ get_id3v20_tags (const gchar *data,
 		{ "COM", "Audio:Comment" },
 #endif /* ENABLE_DETAILED_METADATA */
 		{ "TLE", "Audio:Duration" },
+		{ "TRK", "Audio:TrackNo" },
 		{ NULL, 0 },
 	};
 
@@ -1600,6 +1621,7 @@ get_id3v20_tags (const gchar *data,
 						word = s;
 					} else if (strcmp (tmap[i].text, "TCO") == 0) {
 						gint genre;
+						gchar *improved_genre;
 
 						if (get_genre_number (word, &genre)) {
 							g_free (word);
@@ -1608,9 +1630,12 @@ get_id3v20_tags (const gchar *data,
 							if (g_ascii_strcasecmp (word, "unknown") == 0) {
 								g_free (word);
 								break;
-							} 
-
-							improve_handwritten_genre (word);
+							} else {
+								if ((improved_genre = improve_handwritten_genre (word)) != NULL) {
+									g_free (word);
+									word = improved_genre;
+								}
+							}
 						}
 					} else if (strcmp (tmap[i].text, "TLE") == 0) {
 						guint32 duration;
@@ -1619,8 +1644,15 @@ get_id3v20_tags (const gchar *data,
 						g_free (word);
 						word = g_strdup_printf ("%d", duration/1000);
 						filedata->duration = duration/1000;
-					}	
-					
+					} else if (strcmp (tmap[i].text, "TRK") == 0) {
+						gchar **parts;
+
+						parts = g_strsplit (word, "/", 2);
+						g_free (word);
+
+						word = g_strdup (parts[0]);
+						g_strfreev (parts);
+					}
 					g_hash_table_insert (metadata,
 							     g_strdup (tmap[i].type),
 							     tracker_escape_metadata (word));
@@ -1708,6 +1740,10 @@ parse_id3v24 (const gchar *data,
 			    ((data[12] & 0x7F) << 7) |
 			    ((data[13] & 0x7F) << 0));
 		pos += ehdrSize;
+
+		if (tsize + 10 + ehdrSize > size) {
+			return;
+		}
 	}
 
 	if (unsync) {
@@ -1777,6 +1813,10 @@ parse_id3v23 (const gchar *data,
 
 		pos += 4 + ehdrSize;
 
+		if (tsize + 10 + ehdrSize + padding > size) {
+			return;
+		}
+
 		if (padding < tsize) {
 			tsize -= padding;
 		} else {
